
/* Unix implementation of handleset.h interface, using select() */

#include <evblib/sysdefs/systypes.h>

#include <errno.h>
#include <string.h>

#ifdef HS_DEBUG
#include <stdio.h>
#ifdef __GNUC__
#define hs_debug( args... )	    fprintf( stderr, args )
#else
#define hs_debug( format, ... )	    fprintf( stderr, format, __VA_ARGS__ )
#endif
#else
#define hs_debug( ... )		    /* */
#endif

int handleset_init( HANDLESET_T *hs )
{
    handleset_clear( hs );
    return 0;
}

void handleset_done( HANDLESET_T *hs ) { }

void handleset_clear( HANDLESET_T *hs )
{
    memset( hs, 0, sizeof( *hs ) );
    hs_debug( "hs: cleared handleset %p / %p\n", 
	     &hs->rfds, &hs->wfds );
}

void handleset_addfd( HANDLESET_T *hs, FD_T handle, int for_write )
{
    fd_set *fds;

    if ( handle < 0 ) return;

    if ( handle > hs->highestfd ) hs->highestfd = handle;

    fds = for_write ? &hs->wfds : &hs->rfds;
    FD_SET( handle, fds );

    hs_debug( "hs: added handle %d to handleset %p (%s) - "
		     "highest now %d\n",
	     handle, fds, for_write ? "write" : "read", hs->highestfd );
}

int handleset_hasfd( HANDLESET_T *hs, FD_T handle, int for_write )
{
    int ret;
    fd_set *fds;

    fds = for_write ? &hs->wfds : &hs->rfds;
    ret = handle >= 0 ? FD_ISSET( handle, fds ) : 0;

    hs_debug( "hs: testing handleset %p (%s) for handle %d - "
		     "returning %d\n",
	     fds, for_write ? "write" : "read", handle, ret );
    return ret;
}

void handleset_remfd( HANDLESET_T *hs, FD_T handle, int for_write )
{
    fd_set *fds;

    if ( handle < 0 ) return;

    fds = for_write ? &hs->wfds : &hs->rfds;
    FD_CLR( handle, fds );

    hs_debug( "hs: removed handle %d from handleset %p (%s)\n", 
	     handle, fds, for_write ? "write" : "read" );
}

void handleset_addtimer( HANDLESET_T *hs, TIMER_T *timer )
{
    if ( timerisset( timer ) && timercmp( timer, &hs->firsttimer, < ) )
    {
	hs->firsttimer = *timer;
    }
}

void handleset_addhs( HANDLESET_T *hs, HANDLESET_T *src_hs )
{
    uint32_t *d, *s;
    int cnt;

    handleset_addtimer( hs, &src_hs->firsttimer );

    cnt = sizeof( fd_set ) / sizeof( uint32_t );
    for( d = (uint32_t *)&hs->rfds, s = (uint32_t *)&src_hs->rfds;
	 cnt > 0;
	 cnt-- ) *d++ |= *s++;

    cnt = sizeof( fd_set ) / sizeof( uint32_t );
    for( d = (uint32_t *)&hs->wfds, s = (uint32_t *)&src_hs->wfds;
	 cnt > 0;
	 cnt-- ) *d++ |= *s++;

    if ( src_hs->highestfd > hs->highestfd ) hs->highestfd = src_hs->highestfd;

    hs_debug( "hs: added handleset %p / %p to %p / %p\n", 
	     &src_hs->rfds, &src_hs->wfds, 
	     &hs->rfds, &hs->wfds );
}

void handleset_remhs( HANDLESET_T *hs, HANDLESET_T *src_hs )
{
    uint32_t *d, *s;
    int cnt;

    cnt = sizeof( fd_set ) / sizeof( uint32_t );
    for( d = (uint32_t *)&hs->rfds, s = (uint32_t *)&src_hs->rfds;
	 cnt > 0;
	 cnt-- ) *d++ &= ~*s++;

    cnt = sizeof( fd_set ) / sizeof( uint32_t );
    for( d = (uint32_t *)&hs->wfds, s = (uint32_t *)&src_hs->wfds;
	 cnt > 0;
	 cnt-- ) *d++ &= ~*s++;

    hs_debug( "hs: removed handleset %p / %p from %p / %p\n", 
	     &src_hs->rfds, &src_hs->wfds, 
	     &hs->rfds, &hs->wfds );
}


/* After waiting, the handleset given at the in/out pointer 'io_hs' contains 
   the handles that are signalled. The parameter io_curtime is also an in/out
   pointer; the time value at that address must be set correctly upon entry and
   is set to the current time at exit. Returns -errno on error. */

int handleset_wait( HANDLESET_T *io_hs, TIMER_T *io_curtime )
{
    struct timeval tv, *tvp;
    int ret;

    tvp = 0;
    if ( timerisset( &io_hs->firsttimer ) )
    {
	timer_sub( &tv, &io_hs->firsttimer, io_curtime );
	if ( tv.tv_sec < 0 || tv.tv_usec < 0 ) 
	{
	    tv.tv_sec = 0; tv.tv_usec = 0;
	}
	tvp = &tv;
    }

#ifdef HS_DEBUG
    hs_debug( "hs: waiting for %d fds at %p / %p",
	     io_hs->highestfd + 1, &io_hs->rfds, &io_hs->wfds );
    if ( tvp ) hs_debug( ", timeout %ld.%06ld\n",
		        tvp->tv_sec, tvp->tv_usec );
    else fputc( '\n', stderr );
#endif

    errno = 0;
    ret = select( io_hs->highestfd + 1, &io_hs->rfds, &io_hs->wfds, 0, tvp );
    if ( ret == -1 ) ret = -errno;

    hs_debug( "hs: wait returned %d (%d - %s)\n",
	     ret, errno, strerror( errno ) );

    if ( gettimeofday( io_curtime, 0 ) == -1 ) ret = -errno;
    return ret;
}


/*
 * vim:softtabstop=4:sw=4
 */

